/*****************************************************************************

		Flick FLI-format Animation Viewer v1.0		  19 Dec 1993
		--------------------------------------


This program plays 320x200x8 FLI-format bitmapped animation files on
any ECS or AGA Amiga running OS2.04 or higher.  FLI-format files are
produced by Autodesk Animator and Autodesk 3D Studio on a PC, as well
as by other programs.

The files in this archive may be distributed anywhere provided they are
unmodified and are not sold for profit.

Ownership and copyright of all files remains with the author:

	Peter McGavin, 86 Totara Crescent, Lower Hutt, New Zealand.
	e-mail: peterm@maths.grace.cri.nz

*****************************************************************************/

#include "includes.h"

long __oslibversion = 37;	/* we require at least OS2.0 */

char __stdiowin[] = "CON:20/50/500/130/Flick";
char __stdiov37[] = "/AUTO/CLOSE";

void __asm chunky2planar320x200 (register __a0 UBYTE *chunky_data,
                                 register __a1 PLANEPTR raster,
                                 register __a3 UBYTE *buff0,
                                 register __a4 UBYTE *buff1,
                                 register __a5 struct Library *TimerBase,
                                 register __a6 struct GfxBase *GfxBase);

void __asm chunky2planar640x480 (register __a0 UBYTE *chunky_data,
                                 register __a1 PLANEPTR raster,
                                 register __a3 UBYTE *buff0,
                                 register __a4 UBYTE *buff1,
                                 register __a5 struct Library *TimerBase,
                                 register __a6 struct GfxBase *GfxBase);

#define FLI_256_COLOR  4
#define FLI_DELTA      7
#define FLI_COLOR     11
#define FLI_LC        12
#define FLI_BLACK     13
#define FLI_BRUN      15
#define FLI_COPY      16
#define FLI_MINI      18

struct header {
  ULONG size;
  UWORD magic;
  UWORD frames;
  UWORD width;
  UWORD height;
  UWORD depth;
  UWORD flags;
  UWORD speed;
  ULONG next;
  ULONG frit;
  UBYTE expand[102];
};

struct frameheader {
  ULONG size;
  UWORD magic;
  UWORD chunks;
  UBYTE expand[8];
};

struct chunkheader {
  ULONG size;
  UWORD type;
};

typedef enum mode_type {MODE_COLOUR, MODE_EHB, MODE_GREY};

#define BITMAP_DEPTH 8

PLANEPTR raster[2] = {NULL, NULL};	/* 8 contiguous bitplanes */
struct BitMap screen_bm[2];	/* The displayed bitmap (may be less planes) */
struct BitMap bitmap_bm[2];	/* The full depth-8 bitmap */
struct RastPort rp[2];
UBYTE *buff0 = NULL;		/* CHIP buffer for chunky2planar */
UBYTE *buff1 = NULL;		/* CHIP buffer for chunky2planar */

UWORD __chip emptypointer[] = {
	0x0000, 0x0000,		/* reserved, must be NULL */
	0x0000, 0x0000, 	/* 1 row of image data */
	0x0000, 0x0000};	/* reserved, must be NULL */

struct NewScreen ns = {
		0,0,0,0,0,
		2,1,
		0 /* HIRES | LACE */,
		CUSTOMSCREEN | CUSTOMBITMAP,
		NULL,
		NULL,
		NULL,
		&screen_bm[0]
};

struct NewWindow nw = {
		0,0,			/* Starting corner */
		0,0,			/* Width, height */
		2,1,			/* detail, block pens */
		VANILLAKEY,		/* IDCMP flags */
		ACTIVATE|BORDERLESS,	/* Window flags */
		NULL,			/* Pointer to first gadget */
		NULL,			/* Pointer to checkmark */
		NULL,			/* title */
		NULL,			/* screen pointer */
		NULL,			/* bitmap pointer */
		0,0,0,0,		/* window not sized */
		CUSTOMSCREEN		/* type of screen */
		};

struct Screen *s = NULL;
struct Window *w = NULL;

struct header h;
struct frameheader *fh;
struct chunkheader *ch;

struct Library *TimerBase = NULL;
struct timerequest *timerio = NULL;
ULONG timerclosed = TRUE;
struct EClockVal *time0 = NULL;
struct EClockVal *time1 = NULL;
double micros_per_eclock;	/* Length of EClock tick in microseconds */

char programname[20];
BPTR olddir = NULL;
struct RDArgs *rdargs = NULL;

/****************************************************************************/

void swapw (UWORD *w)
/* Swap the bytes around in a word which may be on an odd byte-boundary */
{
  UBYTE t;

  t = ((UBYTE *)w)[0];
  ((UBYTE *)w)[0] = ((UBYTE *)w)[1];
  ((UBYTE *)w)[1] = t;
}

void swapl (ULONG *l)
/* Swap the bytes around in a longword which may be on an odd byte-boundary */
{
  UBYTE t;

  t = ((UBYTE *)l)[0];
  ((UBYTE *)l)[0] = ((UBYTE *)l)[3];
  ((UBYTE *)l)[3] = t;
  t = ((UBYTE *)l)[1];
  ((UBYTE *)l)[1] = ((UBYTE *)l)[2];
  ((UBYTE *)l)[2] = t;
}

UWORD extractw (UWORD *w)
/* Get a word which may be on an odd byte-boundary */
{
  UBYTE t[2];

  t[0] = ((UBYTE *)w)[0];
  t[1] = ((UBYTE *)w)[1];
  return *(UWORD *)t;
}

ULONG extractl (ULONG *l)
/* Get a longword which may be on an odd byte-boundary */
{
  UBYTE t[4];

  t[0] = ((UBYTE *)l)[0];
  t[1] = ((UBYTE *)l)[1];
  t[2] = ((UBYTE *)l)[2];
  t[3] = ((UBYTE *)l)[3];
  return *(ULONG *)t;
}

/****************************************************************************/

void _STDcleanup (void)
/* This get called automatically by SAS/C 6.3 on any sort of exit condition */
{
  int which;

  if (w != NULL) {
    CloseWindow (w);
    w = NULL;
  }
  if (s != NULL) {
    CloseScreen (s);
    s = NULL;
  }
  if (buff0 != NULL) {
    FreeMem (buff0, h.width * (ULONG)h.height);
    buff0 = NULL;
  }
  if (buff1 != NULL) {
    FreeMem (buff1, h.width * (ULONG)h.height);
    buff1 = NULL;
  }
  for (which = 0; which < 2; which++) {
    if (raster[which] != NULL) {
      FreeRaster (raster[which], h.width, BITMAP_DEPTH * h.height);
      raster[which] = NULL;
    }
  }
  if (time1 != NULL) {
    FreeMem (time1, sizeof(struct EClockVal));
    time1 = NULL;
  }
  if (time0 != NULL) {
    FreeMem (time0, sizeof(struct EClockVal));
    time0 = NULL;
  }
  if (olddir != NULL) {
    CurrentDir (olddir);
    olddir = NULL;
  }
  if (rdargs != NULL) {
    FreeArgs (rdargs);
    rdargs = NULL;
  }
  if (!timerclosed) {
    CloseDevice ((struct IORequest *)timerio);
    TimerBase = NULL;
  }
  if (timerio != NULL) {
    FreeMem (timerio, sizeof(struct timerequest));
    timerio = NULL;
  }
  /* standard libraries are auto-closed here by SAS/C */
}

/****************************************************************************/

void die (char *msg, ...)
/* Exit program with message, return code 10 */
{
  va_list arglist;

  WBenchToFront ();
  va_start (arglist, msg);
  vfprintf (stderr, msg, arglist);
  va_end (arglist);
  Delay (200);
  exit (10);
  /* SAS/C executes _STDcleanup() automatically on exit */
}

/****************************************************************************/

void animate_file (char *fname, BOOL ram, enum mode_type mode, BOOL once)
{
  FILE *f;
  ULONG size, y, class, bytesleftinframe, totalframes, restartpos, l;
  UWORD i, frame, chunk, code;
  UWORD packet, packets, n, m, c, x, line, lines, bsize, depth, which;
  UWORD *viewcolourtable, screen_depth;
  struct IntuiMessage *msg;
  UBYTE *filebuf, *buf, *chunkbuf, *p, *p2, *chunky, *restartptr, *xlate;
  UBYTE (*colourtable)[256][3], pattern[2];
  BOOL going, firstloop;
  ULONG fh_size, ch_size;
  UWORD fh_magic, fh_chunks, ch_type;

  /* open file and read header struct */
  if ((f = fopen (fname, "rb")) == NULL)
    die ("%s: Can't open %s\n", programname, fname);
  if ((fread (&h, sizeof(struct header), 1, f)) != 1)
    die ("%s: Error reading file\n", programname);
  swapl (&h.size);
  swapw (&h.magic);
  swapw (&h.frames);
  swapw (&h.width);
  swapw (&h.height);
  swapw (&h.depth);
  swapw (&h.flags);
  swapw (&h.speed);
  swapl (&h.next);
  swapl (&h.frit);
  printf ("File = %s\n", fname);
  printf ("%lu bytes, %u frames, %ux%ux%u, speed = %u\n", h.size, h.frames,
          h.width, h.height, h.depth, h.speed);
  if (h.magic != 0xaf11 && h.magic != 0xaf12)
    die ("%s: Unrecognised magic number %04x\n", programname, h.magic);
  if ((h.width != 320 || h.height != 200) && (h.width != 640 || h.height != 480))
    die ("%s: Unsupported dimensions\n", programname);

  /* allocate chunky pixels */
  if ((chunky = malloc (h.width * (ULONG)h.height)) == NULL)
    die ("%s: Out of memory\n", programname);
  memset (chunky, 0, h.width * (ULONG)h.height);

  /* allocate 2 rasters (for double-buffering) */
  screen_depth = 8;
  for (which = 0; which < 2; which++) {
    InitBitMap (&screen_bm[which], screen_depth, h.width, h.height); /* Displayed bm */
    InitBitMap (&bitmap_bm[which], BITMAP_DEPTH, h.width, h.height); /* Full depth-8 bm */
    /* Allocate 1 contiguous raster for all 8 planes */
    if ((raster[which] = (PLANEPTR)AllocRaster (h.width,
                                            BITMAP_DEPTH * h.height)) == NULL)
      die ("%s: Out of CHIP memory!\n", programname);
    for (depth = 0; depth < screen_depth; depth++)
      screen_bm[which].Planes[depth] = raster[which] +
                                       depth * RASSIZE (h.width, h.height);
    for (depth = 0; depth < BITMAP_DEPTH; depth++)
      bitmap_bm[which].Planes[depth] = raster[which] +
                                       depth * RASSIZE (h.width, h.height);
    InitRastPort (&rp[which]);
    rp[which].BitMap = &bitmap_bm[which];
    SetRast (&rp[which], 0);
  }

  /* allocate scratch buffers in chip ram for blitter in chunky2planar */
  if ((buff0 = AllocMem (h.width * (ULONG)h.height, MEMF_CHIP)) == NULL)
    die ("%s: Out of CHIP memory!\n", programname);
  if ((buff1 = AllocMem (h.width * (ULONG)h.height, MEMF_CHIP)) == NULL)
    die ("%s: Out of CHIP memory!\n", programname);

  /* if using ram then read in the entire file else allocate header buffers */
  if (ram)
    if ((filebuf = malloc (h.size - sizeof(struct header))) == NULL) {
      printf ("%s: Not enough free contiguous memory to load into RAM\n", programname);
      printf ("       Playing from disk instead\n");
      ram = FALSE;
    } else {
      if ((fread (filebuf, h.size - sizeof(struct header), 1, f)) != 1)
        die ("%s: Error reading file\n", programname);
      buf = filebuf;
    }
  if (!ram) {
    if ((fh = malloc (sizeof(struct frameheader))) == NULL)
      die ("%s: Out of memory\n", programname);
    if ((ch = malloc (sizeof(struct chunkheader))) == NULL)
      die ("%s: Out of memory\n", programname);
  }

  /* open screen with custom bitmap */
  ns.Width = h.width;
  ns.Height = h.height;
  ns.Depth = screen_depth;
  if (h.width > 384)
    ns.ViewModes |= HIRES;
  if (h.height > 283)
    ns.ViewModes |= LACE;
  if (mode == MODE_COLOUR)
    if (GfxBase->LibNode.lib_Version < 39 ||
        (s = OpenScreen(&ns)) == NULL) {
      printf ("%s: Can't open 8 bitplane COLOUR screen\n", programname);
      if (ns.ViewModes & HIRES) {
        printf ("       Using 4 bitplane GREY instead\n");
        mode = MODE_GREY;
      } else {
        printf ("       Using 6 bitplane EHB instead\n");
        mode = MODE_EHB;
      }
      Delay (100);
    }
  if (mode == MODE_EHB || mode == MODE_GREY) {
    if (mode == MODE_EHB) {
      ns.ViewModes |= EXTRA_HALFBRITE;
      screen_depth = 6;
    } else
      screen_depth = 4;
    ns.Depth = screen_depth;
    for (which = 0; which < 2; which++)
      screen_bm[which].Depth = screen_depth;
    if ((s = OpenScreen(&ns)) == NULL)
      die ("%s: Can't open screen!\n", programname);
  }

  /* open a backdrop window (for intuition events) */
  nw.Width = h.width;
  nw.Height = h.height;
  nw.Screen = s;
  nw.MinWidth = h.width;
  nw.MinHeight = h.height;
  nw.MaxWidth = h.width;
  nw.MaxHeight = h.height;
  if ((w = OpenWindow(&nw)) == NULL)
    die ("%s: Can't open window!\n", programname);

  /* turn the mouse pointer off for this window */
  SetPointer (w, emptypointer, 1, 16, 0, 0);

  /* allocate and initialise colour tables and pixel xlate table */
  if ((colourtable = (UBYTE (*)[256][3])malloc (sizeof(*colourtable))) == NULL)
    die ("%s: Out of memory\n", programname);
  memset (colourtable, 0, sizeof(*colourtable));

  if ((xlate = malloc (256)) == NULL)
    die ("%s: Out of memory\n", programname);

  switch (mode) {
    case MODE_EHB:
      if ((viewcolourtable = (UWORD *)malloc ((1 << screen_depth) * sizeof(UWORD))) == NULL)
        die ("%s: Out of memory\n", programname);
      memset (viewcolourtable, 0, (1 << screen_depth) * sizeof(UWORD));
    case MODE_GREY:
      if ((viewcolourtable = (UWORD *)malloc ((1 << screen_depth) * sizeof(UWORD))) == NULL)
        die ("%s: Out of memory\n", programname);
      memset (viewcolourtable, 0, (1 << screen_depth) * sizeof(UWORD));
      for (c = 0; c < 16; c++)
        viewcolourtable[c] = c * 0x0111;
      LoadRGB4 (&s->ViewPort, viewcolourtable, 1 << screen_depth);
      break;
    case MODE_COLOUR:
      if ((viewcolourtable = (UWORD *)malloc (2 * sizeof(UWORD) +
              (1 << screen_depth) * 3 * sizeof(ULONG) + sizeof(UWORD))) == NULL)
        die ("%s: Out of memory\n", programname);
      memset (viewcolourtable, 0, 2 * sizeof(UWORD) +
                       (1 << screen_depth) * 3 * sizeof(ULONG) + sizeof(UWORD));
      viewcolourtable[0] = 1 << screen_depth;
      for (i = 0; i < 256; i++)
        xlate[i] = i;
      break;
  }

  /* initialise loop variables */
  bytesleftinframe = 0; /* diff between framesize and sum of chunksizes in frame */
  which = 0;	/* bitmap index --- flips between 0 and 1 with each frame */
  frame = 0;
  totalframes = 0;
  firstloop = TRUE;
  going = TRUE;

  /* read the start time */
  ReadEClock (time0);

  /* loop for each frame */
  while (going) {

    /* seek to the end of the previous frame */
    if (ram)
      buf += bytesleftinframe;
    else
      if (bytesleftinframe != 0)
        if (fseek (f, bytesleftinframe, SEEK_CUR) == -1)
          die ("%s: Error seeking file\n", programname);

    /* if this is the 2nd frame, save the current file-position for loop */
    if (totalframes == 1)
      if (ram)
        restartptr = buf;
      else
        if ((restartpos = ftell (f)) == -1)
          die ("%s: Error ftelling file\n", programname);

    /* read the frame header */
    if (ram) {
      fh = (struct frameheader *)buf;
      buf += sizeof(struct frameheader);
    } else
      if ((fread (fh, sizeof(struct frameheader), 1, f)) != 1)
        die ("%s: Error reading file\n", programname);
    if (firstloop || !ram) {
      swapl (&fh->size);
      swapw (&fh->magic);
      swapw (&fh->chunks);
    }
    fh_size = extractl (&fh->size);
    fh_magic = extractw (&fh->magic);
    fh_chunks = extractw (&fh->chunks);
    bytesleftinframe = fh_size - sizeof(struct frameheader);

    /* check for and ignore 0x00a1 frames */
    if (fh_magic == 0x00a1) {

      printf ("Ignoring special 0x001a frame\n");

    } else {

      /* render into the non-displayed raster */
      which = 1 - which;

      /* consistency check */
      if (fh_magic != 0xf1fa)
        die ("%s: Unrecognised magic number in frame %04x %lu\n",
                 programname, fh_magic, fh_size);

      /* loop for each chunk */
      for (chunk = 0; chunk < fh_chunks; chunk++) {

        /* read the chunk header */
        if (ram) {
          ch = (struct chunkheader *)buf;
          buf += sizeof(struct chunkheader);
        } else {
          if ((fread (ch, sizeof(struct chunkheader), 1, f)) != 1)
            die ("%s: Error reading file\n", programname);
        }
        if (firstloop || !ram) {
          swapl (&ch->size);
          swapw (&ch->type);
        }
        ch_size = extractl (&ch->size);
        ch_type = extractw (&ch->type);
        bytesleftinframe -= sizeof(struct chunkheader);

        /* correct for odd-length chunks */
        size = (ch_size - sizeof(struct chunkheader) + 1) & ~1;

        /* allocate memory for and read the rest of the chunk */
        if (ram) {
          chunkbuf = buf;
          buf += size;
        } else {
          if ((chunkbuf = malloc(size)) == NULL)
            die ("%s: Out of memory!\n", programname);
          if ((fread (chunkbuf, size, 1, f)) != 1)
            die ("%s: Error reading file\n", programname);
        }
        bytesleftinframe -= size;

        /* uncompress chunk into chunky pixels */
        switch (ch_type) {

          case FLI_DELTA:
            die ("%s: FLI_DELTA chunks aren't debugged yet, sorry\n", programname);
            /* printf ("Got a FLI_DELTA\n"); */
            p = chunkbuf;
            y = 0;
            lines = *p++;
            lines |= (*p++ << 8);
            for (line = 0; line < lines; line++) {
              packets = *p++;
              packets |= (*p++ << 8);
              if (packets > 32767) {
                packets = 65536 - packets;
                y += h.width * packets;
              } else {
                x = 0;
                for (packet = 0; packet < packets; packet++) {
                  x += *p++;
                  bsize = *p++;
                  if (bsize > 127) {
                    bsize = 256 - bsize;
                    if (mode == MODE_COLOUR) {
                      repmem (&chunky[y + x], p, 2, bsize);
                      p += 2;
                    } else {
                      pattern[0] = xlate[*p++];
                      pattern[1] = xlate[*p++];
                      repmem (&chunky[y + x], pattern, 2, bsize);
                    }
                    x += (bsize << 1);
                  } else {
                    bsize <<= 1;
                    if (mode == MODE_COLOUR) {
/*
                      memcpy (&chunky[y + x], p, bsize);
*/
                      CopyMem (p, &chunky[y + x], bsize);
                      p += bsize;
                    } else {
                      p2 = &chunky[y + x];
                      for (i = bsize; i > 0; i--)
                        *p2++ = xlate[*p++];
                    }
                    x += bsize;
                  }
                }
                y += h.width;
              }
            }
            break;

          case FLI_256_COLOR:
          case FLI_COLOR:
            p = chunkbuf;
            packets = *p++;
            packets |= (*p++ << 8);
            c = 0;
            for (packet = 0; packet < packets; packet++) {
              c += *p++;
              n = *p++;
              if (n == 0)
                n = 256;
              for (m = 0; m < n; m++) {
                if (ch_type == FLI_256_COLOR) {
                  (*colourtable)[c][0] = *p++;		/* R */
                  (*colourtable)[c][1] = *p++;		/* G */
                  (*colourtable)[c][2] = *p++;		/* B */
                } else {
                  (*colourtable)[c][0] = *p++ << 2;	/* R */
                  (*colourtable)[c][1] = *p++ << 2;	/* G */
                  (*colourtable)[c][2] = *p++ << 2;	/* B */
                }
                c++;
              }
            }
            switch (mode) {
              case MODE_EHB:
                median_cut (*colourtable, viewcolourtable, xlate);
                LoadRGB4 (&s->ViewPort, viewcolourtable, 32);
                break;
              case MODE_GREY:
                for (c = 0; c < 256; c++)
                  xlate[c] = ((((UWORD)(*colourtable)[c][0]) +
                               ((UWORD)(*colourtable)[c][1]) +
                               ((UWORD)(*colourtable)[c][2])) / 3) >> 4;
                break;
              case MODE_COLOUR:
                for (c = 0; c < (1 << screen_depth); c++) {
                  ((ULONG *)viewcolourtable)[3*c+1] = (*colourtable)[c][0] << 24;
                  ((ULONG *)viewcolourtable)[3*c+2] = (*colourtable)[c][1] << 24;
                  ((ULONG *)viewcolourtable)[3*c+3] = (*colourtable)[c][2] << 24;
                }
                LoadRGB32 (&s->ViewPort, (ULONG *)viewcolourtable);
                break;
            }
            break;            

          case FLI_LC:
            p = chunkbuf;
            y = *p++;
            y |= (*p++ << 8);
            y *= h.width;
            lines = *p++;
            lines |= (*p++ << 8);
            for (line = 0; line < lines; line++) {
              packets = *p++;
              x = 0;
              for (packet = 0; packet < packets; packet++) {
                x += *p++;
                bsize = *p++;
                if (bsize > 127) {
                  bsize = 256 - bsize;
                  memset (&chunky[y + x], xlate[*p++], bsize);
                  x += bsize;
                } else {
                  if (mode == MODE_COLOUR) {
                    memcpy (&chunky[y + x], p, bsize);
/*
                    CopyMem (p, &chunky[y + x], bsize);
*/
                    p += bsize;
                  } else {
                    p2 = &chunky[y + x];
                    for (i = bsize; i > 0; i--)
                      *p2++ = xlate[*p++];
                  }
                  x += bsize;
                }
              }
              y += h.width;
            }
            break;

          case FLI_BLACK:
            memset (chunky, xlate[0], h.width * (ULONG)h.height);
            break;

          case FLI_BRUN:
            p = chunkbuf;
            y = 0;
            for (line = 0; line < h.height; line++) {
              packets = *p++;
              x = 0;
              for (packet = 0; packet < packets; packet++) {
                bsize = *p++;
                if (bsize <= 127) {
                  memset (&chunky[y + x], xlate[*p++], bsize);
                  x += bsize;
                } else {
                  bsize = 256 - bsize;
                  if (mode == MODE_COLOUR) {
                    memcpy (&chunky[y + x], p, bsize);
/*
                    CopyMem (p, &chunky[y + x], bsize);
*/
                    p += bsize;
                  } else {
                    p2 = &chunky[y + x];
                    for (i = bsize; i > 0; i--)
                      *p2++ = xlate[*p++];
                  }
                  x += bsize;
                }
              }
              y += h.width;
            }
            break;

          case FLI_COPY:
            if (mode == MODE_COLOUR)
/*
              memcpy (chunky, chunkbuf, h.width * (ULONG)h.height);
*/
              CopyMemQuick (chunkbuf, chunky, h.width * (ULONG)h.height);
            else {
              p2 = chunky;
              for (l = h.width * (ULONG)h.height; l > 0; l--)
                *p2++ = xlate[*p++];
            }
            break;

          case FLI_MINI:
            break;

          default:
            die ("%s: Unrecognised chunk type %04x\n", programname, ch_type);
            break;
        } /* end of switch for each chunk type */

        /* deallocate chunk ram */
        if (!ram)
          free (chunkbuf);

      } /* end of loop for each chunk within frame */

      /* convert from chunky pixels to (hidden) planar raster[which] */
      if (h.width == 320 && h.height == 200)
        chunky2planar320x200 (chunky, raster[which], buff0, buff1, TimerBase, GfxBase);
      else if (h.width == 640 && h.height == 480)
        chunky2planar640x480 (chunky, raster[which], buff0, buff1, TimerBase, GfxBase);
      else
        die ("%s: Unsupported resolution\n", programname);

      /* make the new raster visible */
      s->ViewPort.RasInfo->BitMap = &screen_bm[which];
      MakeScreen (s);
      RethinkDisplay ();

      /* check for CTRL/C or BREAK */
      if (SetSignal (0, 0) & SIGBREAKF_CTRL_C) {
        SetSignal (0, SIGBREAKF_CTRL_C);
        going = FALSE;
      }

      /* check for intuition messages */
      while ((msg = (struct IntuiMessage *)GetMsg (w->UserPort)) != NULL) {
        class = msg->Class;
        code = msg->Code;
        ReplyMsg ((struct Message *)msg);
        if (class == IDCMP_VANILLAKEY)
          switch (code) {
            case 0x03: /* CTRL/C */
            case 0x1b: /* ESC */
            case 0x51: /* q */
            case 0x71: /* Q */
              going = FALSE;
            default:
              break;
          }
      }

    } /* end of if frame magic != 0x00a1 */

    /* seek back to frame 1 (the 2nd frame) when we get to the last frame */
    if (frame++ == h.frames)
      if (once)
        going = FALSE;
      else {
        firstloop = FALSE;
        frame = 1;
        if (ram)
          buf = restartptr;
        else
          if (fseek (f, restartpos, SEEK_SET) == -1)
            die ("%s: Error seeking file\n", programname);
        bytesleftinframe = 0;
      }

    totalframes++;

  } /* end of loop for each frame */

  /* find out and display how long it took */
  ReadEClock (time1);
  printf ("%s: Frames per second = %5.2lf\n", programname, 1000000.0 * totalframes /
                      ((time1->ev_lo - time0->ev_lo) * micros_per_eclock));

  /* close files and free memory */
  fclose (f);
  free (colourtable);
  free (viewcolourtable);
  free (xlate);
  if (ram)
    free (filebuf);
  else {
    free (ch);
    free (fh);
  }
  free (chunky);
  CloseWindow (w);
  w = NULL;
  CloseScreen (s);
  s = NULL;
  FreeMem (buff0, h.width * (ULONG)h.height);
  buff0 = NULL;
  FreeMem (buff1, h.width * (ULONG)h.height);
  buff1 = NULL;
  for (which = 0; which < 2; which++) {
    FreeRaster (raster[which], h.width, BITMAP_DEPTH * h.height);
    raster[which] = NULL;
  }
}

/****************************************************************************/

BOOL ram = TRUE;
enum mode_type mode = MODE_COLOUR;
BOOL once = FALSE;

struct Library *AslBase = NULL;

struct TagItem frtags[] = {
  {ASL_Hail,		0},
  {ASL_Pattern,		(ULONG)"#?.fl[ic]"},
  {ASL_OKText,		(ULONG)"Play"},
  {ASL_CancelText,	(ULONG)"Cancel"},
  {ASL_FuncFlags,	FILF_PATGAD},
  {TAG_DONE,		0}
};

char resultstring[128];

void filerequestloop (void)
{
  struct FileRequester *fr;

  if ((AslBase = OpenLibrary ("asl.library", 37L)) != NULL) {
    frtags[0].ti_Data = (ULONG)programname;
    if (AslBase->lib_Version < 39)
      frtags[1].ti_Data = (ULONG)"#?.fl?";
    if ((fr = (struct FileRequester *)AllocAslRequest
                                          (ASL_FileRequest, frtags)) != NULL) {
      while (AslRequest (fr, NULL)) {
        strcpy (resultstring, fr->rf_Dir);
        AddPart (resultstring, fr->rf_File, 128);
        animate_file (resultstring, ram, mode, once);
      }
      FreeAslRequest (fr);
    }
    CloseLibrary (AslBase);
  }
}

/****************************************************************************/

LONG argarray[9] = {0,0,0,0,0,0,0,0,0};

int main (int argc, char *argv[])
{
  struct WBStartup *argmsg;
  struct WBArg *wb_arg;
  char **fnames;
  UWORD ktr;

  /* standard libraries are auto-opened here by SAS/C */

  /* find out name of program */
  GetProgramName (programname, 19);

  /* timer stuff */
  if ((timerio = (struct timerequest *)AllocMem (sizeof(struct timerequest),
                                           MEMF_CLEAR | MEMF_PUBLIC)) == NULL)
    die ("%s: Out of memory\n", programname);
  if (timerclosed = OpenDevice (TIMERNAME, UNIT_MICROHZ,
                                (struct IORequest *)timerio, 0))
    die ("%s: Can't open timer.device!\n", programname);
  TimerBase = (struct Library *)timerio->tr_node.io_Device;
  if ((time0 = (struct EClockVal *)AllocMem (sizeof(struct EClockVal),
                                             MEMF_CLEAR | MEMF_PUBLIC)) == NULL)
    die ("%s: Out of memory\n", programname);
  if ((time1 = (struct EClockVal *)AllocMem (sizeof(struct EClockVal),
                                             MEMF_CLEAR | MEMF_PUBLIC)) == NULL)
    die ("%s: Out of memory\n", programname);
  micros_per_eclock = 1000000.0 / (double)ReadEClock (time0);

  /* parse workbench message or commandline */
  if (argc == 0) {
    argmsg = (struct WBStartup *)argv;
    wb_arg = argmsg->sm_ArgList;
    strcpy (programname, wb_arg->wa_Name);
    if (argmsg->sm_NumArgs <= 1)
      filerequestloop ();
    else {
      wb_arg++;
      for (ktr = 1; ktr < argmsg->sm_NumArgs; ktr++, wb_arg++)
        if (wb_arg->wa_Lock != NULL) {
          olddir = CurrentDir (wb_arg->wa_Lock);
          animate_file (wb_arg->wa_Name, ram, mode, once);
          CurrentDir (olddir);
          olddir = NULL;
        } else
          animate_file (wb_arg->wa_Name, ram, mode, once);
      }
  } else {
    if ((rdargs = ReadArgs ("FILE/M,DISK/S,RAM/S,ONCE/S,COLOUR/S,COLOR/S,EHB/S,GREY/S,GRAY/S,",
                            argarray, NULL)) != NULL) {
      if (argarray[1])
        ram = FALSE;
      if (argarray[2])
        ram = TRUE;
      if (argarray[3])
        once = TRUE;
      if (argarray[4] || argarray[5])
        mode = MODE_COLOUR;
      if (argarray[6])
        mode = MODE_EHB;
      if (argarray[7] || argarray[8])
        mode = MODE_GREY;
      fnames = (char **)argarray[0];
      if (fnames == NULL || *fnames == NULL)
        filerequestloop ();
      else
        while (*fnames != NULL)
          animate_file (*fnames++, ram, mode, once);
      FreeArgs (rdargs);
      rdargs = NULL;
    }
  }

  /* Let the user see how fast it went before the window closes */
  if (argc == 0)
    Delay (200);

  /* exit program with success code */
  return (0);

  /* SAS/C executes _STDcleanup() automatically on exit or break */

}

/****************************************************************************/
